@@ -42,6 +42,7 @@ import signal
 import socket
 import json
 import traceback
+import time
 
 from LogMessage import openLog, log
 
@@ -55,8 +56,8 @@ LOGPATH            = os.path.join(MONITORLOGLOCATION,MONITORLOGFILENAME)
 HISTORYFILENAME    = "monitorPBS.history"
 HISTORYFILEPATH    = os.path.join(MONITORROOT,HISTORYFILENAME)
 
-SLEEPTIME       = 30
-PAUSETIME       = 5.
+SLEEPTIME       = 60
+PAUSETIME       = 15.
 MAXIMUMIDLETIME = 30*60
 
 
@@ -78,10 +79,14 @@ class QueueMonitor:
       self.maximumConsecutiveEmptyQueues = maximumIdleTime/sleepTime
 
       self.historyFile = None
-      self.updates     = []
       self.activeJobs  = {}
       self.bufferSize  = 4096
 
+      self.updates                    = {}
+      self.updateBucket               = 0
+      self.updates[self.updateBucket] = []
+      self.updatesPerBucket           = 50
+
       signal.signal(signal.SIGINT,self.sigINT_handler)
       signal.signal(signal.SIGHUP,self.sigHUP_handler)
       signal.signal(signal.SIGQUIT,self.sigQUIT_handler)
@@ -155,9 +160,6 @@ class QueueMonitor:
                                                                                   self.activeJobs[jobId]['stage'], \
                                                                                   self.activeJobs[jobId]['queue']))
       self.historyFile.flush()
-#     self.updates.append(str(jobId) + ' ' + self.activeJobs[jobId][0] + ' ' + \
-#                                            self.activeJobs[jobId][1] + ' ' + \
-#                                            self.activeJobs[jobId][2])
       jobState = {'jobId':jobId,
                   'status':self.activeJobs[jobId]['status'],
                   'stage':self.activeJobs[jobId]['stage'],
@@ -178,7 +180,12 @@ class QueueMonitor:
                   text = ""
             jobState['tailFiles'][tailFile] = text
 
-      self.updates.append(jobState)
+#     log(str(jobState))
+      if len(self.updates[self.updateBucket]) == self.updatesPerBucket:
+         self.updateBucket += 1
+         self.updates[self.updateBucket] = []
+
+      self.updates[self.updateBucket].append(jobState)
 
 
    def loadHistory(self):
@@ -207,9 +214,9 @@ class QueueMonitor:
       self.openHistory('w')
       if self.historyFile:
          for activeJob in self.activeJobs:
-            self.historyFile.write("%s:%s %s %s %s\n" % (self.siteDesignator,str(activeJob), \
-                                                                             self.activeJobs[activeJob]['status'], \
-                                                                             self.activeJobs[activeJob]['stage'], \
+            self.historyFile.write("%s:%s %s %s %s\n" % (self.siteDesignator,str(activeJob),
+                                                                             self.activeJobs[activeJob]['status'],
+                                                                             self.activeJobs[activeJob]['stage'],
                                                                              self.activeJobs[activeJob]['queue']))
          self.historyFile.close()
          self.historyFile = None
@@ -348,23 +355,25 @@ class QueueMonitor:
                   nLines,
                   lastEndPosition=0):
       text        = ""
+      nTextLines  = 0
       endPosition = lastEndPosition
       bufsize = 2048
       if os.path.exists(tailPath):
          fileSize = os.stat(tailPath).st_size
-         if lastEndPosition+1 <= fileSize:
+         if fileSize > lastEndPosition:
             try:
                fpTail = open(tailPath,'r')
                try:
                   iter = 0
                   while True:
                      iter += 1
-                     location = max(fileSize-bufsize*iter,lastEndPosition+1)
+                     location = max(fileSize-bufsize*iter,lastEndPosition)
                      fpTail.seek(location)
                      data = fpTail.readlines()
                      endPosition = fpTail.tell()
-                     if len(data) > nLines or location == lastEndPosition+1:
+                     if len(data) > nLines or location == lastEndPosition:
                         text = ''.join(data[-nLines:])
+                        nTextLines += min(len(data),nLines)
                         if lastEndPosition > 0 and location > lastEndPosition+1:
                            text = "...\n" + text
                         del data
@@ -376,6 +385,9 @@ class QueueMonitor:
             except (IOError,OSError):
                log("%s could not be opened" % (tailPath))
 
+      if nTextLines > 0:
+         log("tailed %d lines of %s" % (nTextLines,tailPath))
+
       return(text,endPosition)
 
 
@@ -384,141 +396,210 @@ class QueueMonitor:
       consecutiveEmptyQueues = 0
       lastReportedActiveJobCount = -1
 
-      toCheck = []
-      toCheck.append(sys.stdin.fileno())
+      readers = []
+      readers.append(sys.stdin.fileno())
+      writers = []
+      writers.append(sys.stdout.fileno())
+
       while True:
          activeJobCount = len(self.activeJobs)
          if activeJobCount != lastReportedActiveJobCount:
             log("%d monitored jobs" % (activeJobCount))
          lastReportedActiveJobCount = activeJobCount
 
-         self.updates  = []
+         self.updates                    = {}
+         self.updateBucket               = 0
+         self.updates[self.updateBucket] = []
+
          currentJobs   = {}
          completedJobs = []
 
-         delayTime = 0
-         while delayTime <= self.sleepTime:
+         startReadTime = time.time()
+         while time.time()-startReadTime <= self.sleepTime:
             if os.getppid() == 1:
                os.kill(os.getpid(),signal.SIGTERM)
 
-            ready = select.select(toCheck,[],[],self.pauseTime) # wait for input
-            if sys.stdin.fileno() in ready[0]:
-#              newJob = sys.stdin.readline().strip()
-               newJob = ""
+            readyReaders,readyWriters,readyExceptions = select.select(readers,[],[],self.pauseTime) # wait for input
+            if sys.stdin.fileno() in readyReaders:
                message = sys.stdin.readline()
-               if message != '':
-                  try:
-                     centralMessage = json.loads(message)
-                  except:
-                     log(traceback.format_exc())
-                  else:
-                     if centralMessage['messageType'] == 'newJobId':
-                        newJob = centralMessage['remoteJobId']
-                  if newJob != "":
-                     if not newJob in self.activeJobs:
-                        self.activeJobs[newJob] = {'jobId':newJob,
-                                                   'status':'N',
-                                                   'stage':'Job',
-                                                   'queue':'?',
-                                                   'jobWorkDirectory':centralMessage.get('jobWorkDirectory','?'),
-                                                   'localJobId':centralMessage.get('localJobId','?'),
-                                                   'instanceId':centralMessage.get('instanceId','?'),
-                                                   'runName':centralMessage.get('runName','?'),
-                                                   'tailFiles':centralMessage.get('tailFiles',[])}
-                        jobWorkDirectory = os.path.expandvars(os.path.expanduser(self.activeJobs[newJob]['jobWorkDirectory']))
-                        if 'tailFiles' in self.activeJobs[newJob]:
-                           if 'runName' in self.activeJobs[newJob]:
-                              runName = self.activeJobs[newJob]['runName']
-                              if 'instanceId' in self.activeJobs[newJob]:
-                                 instanceId = self.activeJobs[newJob]['instanceId']
-                                 if "#STDOUT#" in self.activeJobs[newJob]['tailFiles']:
-                                    stdFile = "%s_%s.stdout" % (runName,instanceId)
-                                    self.activeJobs[newJob]['tailFiles']["#STDOUT#"]['path'] = os.path.join(jobWorkDirectory,
-                                                                                                            stdFile)
-                                    self.activeJobs[newJob]['tailFiles']["#STDOUT#"]['endPosition'] = 0
-                                 if "#STDERR#" in self.activeJobs[newJob]['tailFiles']:
-                                    stdFile = "%s_%s.stderr" % (runName,instanceId)
-                                    self.activeJobs[newJob]['tailFiles']["#STDERR#"]['path'] = os.path.join(jobWorkDirectory,
-                                                                                                            stdFile)
-                                    self.activeJobs[newJob]['tailFiles']["#STDERR#"]['endPosition'] = 0
-                           for tailFile in self.activeJobs[newJob]['tailFiles']:
-                              if tailFile != "#STDOUT#" and tailFile != "#STDERR#":
-                                 self.activeJobs[newJob]['tailFiles'][tailFile]['path'] = os.path.join(jobWorkDirectory,
-                                                                                                       tailFile)
-                                 self.activeJobs[newJob]['tailFiles'][tailFile]['endPosition'] = 0
-                        self.recordHistory(newJob)
-                        self.activeJobs[newJob]['status'] = 'n'
-                     consecutiveEmptyQueues = 0
+               if message != "":
+                  nNewJobs    = 0
+                  nOrphanJobs = 0
+                  while message != "":
+                     try:
+                        centralMessage = json.loads(message)
+                     except:
+                        log(traceback.format_exc())
+                     else:
+#                       log("READ: " + str(centralMessage))
+                        if   centralMessage['messageType'] == 'newJobId':
+                           newJob = centralMessage['remoteJobId']
+#                          log("newJobId %s" % (newJob))
+                           nNewJobs += 1
+                           if not newJob in self.activeJobs:
+                              self.activeJobs[newJob] = {'jobId':newJob,
+                                                         'status':'N',
+                                                         'stage':'Job',
+                                                         'queue':'?',
+                                                         'jobWorkDirectory':centralMessage.get('jobWorkDirectory','?'),
+                                                         'localJobId':centralMessage.get('localJobId','?'),
+                                                         'instanceId':centralMessage.get('instanceId','?'),
+                                                         'runName':centralMessage.get('runName','?'),
+                                                         'tailFiles':centralMessage.get('tailFiles',{})}
+                              jobWorkDirectory = os.path.expandvars(os.path.expanduser(self.activeJobs[newJob]['jobWorkDirectory']))
+                              if 'tailFiles' in self.activeJobs[newJob]:
+                                 if 'runName' in self.activeJobs[newJob]:
+                                    runName = self.activeJobs[newJob]['runName']
+                                    if 'instanceId' in self.activeJobs[newJob]:
+                                       instanceId = self.activeJobs[newJob]['instanceId']
+                                       if "#STDOUT#" in self.activeJobs[newJob]['tailFiles']:
+                                          stdFile = "%s_%s.stdout" % (runName,instanceId)
+                                          self.activeJobs[newJob]['tailFiles']["#STDOUT#"]['path'] = os.path.join(jobWorkDirectory,
+                                                                                                                  stdFile)
+                                          self.activeJobs[newJob]['tailFiles']["#STDOUT#"]['endPosition'] = 0
+                                       if "#STDERR#" in self.activeJobs[newJob]['tailFiles']:
+                                          stdFile = "%s_%s.stderr" % (runName,instanceId)
+                                          self.activeJobs[newJob]['tailFiles']["#STDERR#"]['path'] = os.path.join(jobWorkDirectory,
+                                                                                                                  stdFile)
+                                          self.activeJobs[newJob]['tailFiles']["#STDERR#"]['endPosition'] = 0
+                                 for tailFile in self.activeJobs[newJob]['tailFiles']:
+                                    if tailFile != "#STDOUT#" and tailFile != "#STDERR#":
+                                       self.activeJobs[newJob]['tailFiles'][tailFile]['path'] = os.path.join(jobWorkDirectory,
+                                                                                                             tailFile)
+                                       self.activeJobs[newJob]['tailFiles'][tailFile]['endPosition'] = 0
+                              self.recordHistory(newJob)
+                              self.activeJobs[newJob]['status'] = 'n'
+                           consecutiveEmptyQueues = 0
+                        elif centralMessage['messageType'] == 'orphanJobId':
+                           orphanJob = centralMessage['remoteJobId']
+#                          log("orphanJobId %s" % (orphanJob))
+                           nOrphanJobs += 1
+                           if not orphanJob in self.activeJobs:
+                              self.activeJobs[orphanJob] = {'jobId':orphanJob,
+                                                            'status':'N',
+                                                            'stage':'Job',
+                                                            'queue':'?',
+                                                            'jobWorkDirectory':'?',
+                                                            'localJobId':centralMessage.get('localJobId','?'),
+                                                            'instanceId':centralMessage.get('instanceId','?'),
+                                                            'runName':centralMessage.get('runName','?'),
+                                                            'tailFiles':{}}
+                           self.activeJobs[orphanJob]['status'] = 'o'
+                        elif centralMessage['messageType'] == 'pipeFlusher':
+                           break
+                        else:
+                           log(centralMessage)
+
+                     message = sys.stdin.readline()
+
+                  if nNewJobs > 0:
+                     log("%d newJobs" % (nNewJobs))
+                  if nOrphanJobs > 0:
+                     log("%d orphanJobs" % (nOrphanJobs))
                else:
                   self.cleanup()
                   log("%s empty job monitor stopped" % (self.siteDesignator))
                   sys.exit(0)
-            delayTime += self.pauseTime
 
-         errStatus,qstatOutput,qstatError = self.executeQstatCommand(self.qstatCommand)
-         if errStatus == 0:
-            jobs = self.filterQstat(qstatOutput.splitlines())
-            for job in jobs:
-               jobId = job['jobId']
-               currentJobs[jobId] = job
-#              currentJobs[jobId] = (status,stage,queue)
+         try:
+            readyReaders,readyWriters,readyExceptions = select.select([],writers,[],self.pauseTime) # wait for input
+         except:
+            readyWriters = []
+         if sys.stdout.fileno() in readyWriters:
+            errStatus,qstatOutput,qstatError = self.executeQstatCommand(self.qstatCommand)
+            if errStatus == 0:
+               jobs = self.filterQstat(qstatOutput.splitlines())
+               for job in jobs:
+                  jobId = job['jobId']
+                  currentJobs[jobId] = job
 
-            if len(currentJobs) == 0:
-               consecutiveEmptyQueues += 1
+               if len(currentJobs) == 0:
+                  consecutiveEmptyQueues += 1
+               else:
+                  consecutiveEmptyQueues = 0
+
+               for activeJob in self.activeJobs:
+                  if self.activeJobs[activeJob]['status'] == 'n':
+                     self.activeJobs[activeJob]['status'] = 'N'
+                     self.activeJobs[activeJob]['stage']  = 'Job'
+                     self.activeJobs[activeJob]['queue']  = '?'
+                  else:
+                     if not activeJob in currentJobs:
+                        self.activeJobs[activeJob]['status'] = 'D'
+                        self.recordHistory(activeJob)
+                        completedJobs.append(activeJob)
+
+               for currentJob in currentJobs:
+                  if not currentJob in self.activeJobs:
+#                    log("recordHistory: currentJob[%s] not in activeJobs" % (currentJob))
+                     self.activeJobs[currentJob] = currentJobs[currentJob]
+                     self.recordHistory(currentJob)
+                  else:
+                     somethingChanged = False
+                     for key in currentJobs[currentJob]:
+                        if currentJobs[currentJob][key] != self.activeJobs[currentJob][key]:
+                           somethingChanged = True
+                     if somethingChanged:
+#                       log("recordHistory: currentJob[%s] different than activeJob" % (currentJob))
+                        for key in currentJobs[currentJob]:
+                           self.activeJobs[currentJob][key] = currentJobs[currentJob][key]
+                        self.recordHistory(currentJob)
+                     elif 'tailFiles' in self.activeJobs[currentJob]:
+                        for tailFile in self.activeJobs[currentJob]['tailFiles']:
+                           tailPath = self.activeJobs[currentJob]['tailFiles'][tailFile]['path']
+                           if os.path.exists(tailPath):
+                              fileSize = os.stat(tailPath).st_size
+                              if fileSize > self.activeJobs[currentJob]['tailFiles'][tailFile]['endPosition']:
+#                                log("recordHistory: change in tailFiles")
+                                 self.recordHistory(currentJob)
+                                 break
+
+                  if self.activeJobs[currentJob]['status'] == 'D':
+                     completedJobs.append(currentJob)
+
+               for completedJob in completedJobs:
+                  del self.activeJobs[completedJob]
+
+               del currentJobs
+               del completedJobs
+
+               if len(self.updates) > 0:
+                  for bucket in self.updates:
+                     nUpdates = len(self.updates[bucket])
+                     if nUpdates > 0:
+                        siteMessage = {'messageType':'siteUpdate','siteDesignator':self.siteDesignator,
+                                       'nJobStates':nUpdates,'jobStates':self.updates[bucket]}
+                        try:
+#                          log("WRITE: " + str(siteMessage))
+                           sys.stdout.write(json.dumps(siteMessage) + '\n')
+                        except:
+                           log("Site update of %d jobs failed" % (nUpdates))
+                        else:
+                           log("Site updated %d jobs" % (nUpdates))
+                           sys.stdout.flush()
+
+                        siteMessage = {'messageType':'pipeFlusher'}
+                        try:
+                           sys.stdout.write(json.dumps(siteMessage) + '\n')
+                        except:
+                           log("Site pipe flush failed")
+                        else:
+#                          log("Site pipe flushed")
+                           sys.stdout.flush()
+               else:
+                  siteMessage = {'messageType':'pipeFlusher'}
+                  try:
+#                    log("WRITE: " + str(siteMessage))
+                     sys.stdout.write(json.dumps(siteMessage) + '\n')
+                  except:
+                     log("Site pipe flush failed")
+                  else:
+#                    log("Site pipe flushed")
+                     sys.stdout.flush()
+
+               del self.updates
             else:
-               consecutiveEmptyQueues = 0
-
-            for activeJob in self.activeJobs:
-               if self.activeJobs[activeJob]['status'] == 'n':
-                  self.activeJobs[activeJob]['status'] = 'N'
-                  self.activeJobs[activeJob]['stage']  = 'Job'
-                  self.activeJobs[activeJob]['queue']  = '?'
-               else:
-                  if not activeJob in currentJobs:
-                     self.activeJobs[activeJob]['status'] = 'D'
-                     self.recordHistory(activeJob)
-                     completedJobs.append(activeJob)
-
-            for currentJob in currentJobs:
-               if   not currentJob in self.activeJobs:
-                  self.activeJobs[currentJob] = currentJobs[currentJob]
-                  self.recordHistory(currentJob)
-               elif currentJobs[currentJob] != self.activeJobs[currentJob]:
-                  for key in currentJobs[currentJob]:
-                     self.activeJobs[currentJob][key] = currentJobs[currentJob][key]
-                  self.recordHistory(currentJob)
-               elif 'tailFiles' in self.activeJobs[currentJob]:
-                  for tailFile in self.activeJobs[currentJob]['tailFiles']:
-                     tailPath = self.activeJobs[currentJob]['tailFiles'][tailFile]['path']
-                     if os.path.exists(tailPath):
-                        fileSize = os.stat(tailPath).st_size
-                        if fileSize > self.activeJobs[currentJob]['tailFiles'][tailFile]['endPosition']:
-                           self.recordHistory(currentJob)
-                           break
-
-               if self.activeJobs[currentJob]['status'] == 'D':
-                  completedJobs.append(currentJob)
-
-            for completedJob in completedJobs:
-               del self.activeJobs[completedJob]
-
-            del currentJobs
-            del completedJobs
-
-            if len(self.updates) > 0:
-#              updateMessage = str(len(self.updates)) + ' ' + self.siteDesignator + ':' + ':'.join(self.updates)
-#              sys.stdout.write("%s\n" % (updateMessage))
-#              sys.stdout.flush()
-               siteMessage = {'messageType':'siteUpdate','siteDesignator':self.siteDesignator,
-                              'nJobStates':len(self.updates),'jobStates':self.updates}
-               try:
-                  sys.stdout.write(json.dumps(siteMessage) + '\n')
-               except:
-                  log("Site update failed")
-               else:
-                  sys.stdout.flush()
-
-            del self.updates
+               log("Error %d in %s command:\n%s" % (errStatus,self.qstatCommand,qstatError))
 
             if self.historyFile:
                self.historyFile.close()
@@ -530,8 +611,6 @@ class QueueMonitor:
                self.cleanup()
                log("%s idle monitor stopped" % (self.siteDesignator))
                sys.exit(0)
-         else:
-            log("Error %d in %s command:\n%s" % (errStatus,self.qstatCommand,qstatError))
 
 
 if __name__ == '__main__':
@@ -544,6 +623,7 @@ if __name__ == '__main__':
       else:
          if sys.argv[1] != SITEDESIGNATOR:
             log("Site designators do not match.\n   internal name = %s\n   external name = %s" % (SITEDESIGNATOR,sys.argv[1]))
+            siteDesignator = ""
          else:
             siteDesignator = sys.argv[1]
    else:
@@ -554,8 +634,8 @@ if __name__ == '__main__':
 
    log("%s monitor started on %s" % (siteDesignator,socket.gethostname()))
 
-   __queueMonitor__ = QueueMonitor(siteDesignator,USERNAME,QSTATCOMMAND,HISTORYFILEPATH,
-                                   SLEEPTIME,PAUSETIME,MAXIMUMIDLETIME)
+   __queueMonitor__ = QueueMonitor(siteDesignator,USERNAME,QSTATCOMMAND,
+                                   HISTORYFILEPATH,SLEEPTIME,PAUSETIME,MAXIMUMIDLETIME)
 
    __queueMonitor__.loadHistory()
    __queueMonitor__.saveHistory()
