@@ -378,6 +378,8 @@ class MonitorJob:
             if   readyReader in listeningSocket:
                if not self.jobMonitor.acceptConnection():
                   self.logger.log(logging.ERROR,getLogMessage("Connection failed."))
+               else:
+                  self.logger.log(logging.DEBUG,getLogMessage("Connection accepted."))
             elif readyReader in activeReaders:
                self.jobMonitor.receiveMessage(readyReader)
 
