@@ -129,31 +129,34 @@ def monitorJobR(listenURI):
                   reportedJobs[status] = []
                reportedJobs[status].append(jobId)
 
-            for status in reportedJobs.keys():
-               if not status in ignoreStatus:
-                  statusHeaderPrinted = False
+            try:
+               for status in reportedJobs.keys():
+                  if not status in ignoreStatus:
+                     statusHeaderPrinted = False
 
-                  jobIds = reportedJobs[status]
-                  if len(jobIds) > 0:
-                     if not whenHeaderPrinted:
-                        print "Current time              %s" % (time.ctime())
-                        print "Last time status reported %s" % (time.ctime(report['maxLastReportTime']))
-                        whenHeaderPrinted = True
-                     if not siteDesignatorHeaderPrinted:
-                        lastReportTime = siteDesignators[siteDesignator]['lastReportTime']
-                        print "\nReport for monitor %s" % (siteDesignator)
-                        print "   Last time status reported %s" % (time.ctime(float(lastReportTime)))
-                        siteDesignatorHeaderPrinted = True
-                     if not queueHeaderPrinted:
-                        print "   Queue %s" % (queue)
-                        queueHeaderPrinted = True
-                     if not statusHeaderPrinted:
-                        print "      Jobs in status: %s" % (status)
-                        statusHeaderPrinted = True
+                     jobIds = reportedJobs[status]
+                     if len(jobIds) > 0:
+                        if not whenHeaderPrinted:
+                           print "Current time              %s" % (time.ctime())
+                           print "Last time status reported %s" % (time.ctime(report['maxLastReportTime']))
+                           whenHeaderPrinted = True
+                        if not siteDesignatorHeaderPrinted:
+                           lastReportTime = siteDesignators[siteDesignator]['lastReportTime']
+                           print "\nReport for monitor %s" % (siteDesignator)
+                           print "   Last time status reported %s" % (time.ctime(float(lastReportTime)))
+                           siteDesignatorHeaderPrinted = True
+                        if not queueHeaderPrinted:
+                           print "   Queue %s" % (queue)
+                           queueHeaderPrinted = True
+                        if not statusHeaderPrinted:
+                           print "      Jobs in status: %s" % (status)
+                           statusHeaderPrinted = True
 
-                     natsort(jobIds)
-                     for jobId in jobIds:
-                        print "         %s" % (jobId)
+                        natsort(jobIds)
+                        for jobId in jobIds:
+                           print "         %s" % (jobId)
+            except IOError:
+               pass
 
             del reportedJobs
 
